################################################################################
### Yggdrasil.pm
### Generates a website from a directory tree of content and templates
###
### Copyright (C) 2003 Morten Wulff <wulff@ratatosk.net>
###
### This program is free software; you can redistribute it and/or
### modify it under the terms of the GNU General Public License
### as published by the Free Software Foundation; either version 2
### of the License, or (at your option) any later version.
###
### This program is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with this program; if not, write to the Free Software
### Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
################################################################################

### REVISIONS
###	3.0.0 : 2003-04-27 : Morten Wulff <wulff@ratatosk.net>
###			Modularized rewrite of version 2.2.0.
###	3.1.0 :	2003-04-29 : Morten Wulff <wulff@ratatosk.net>
###			Added support for the SP SGML System. Download SP
###			(http://www.jclark.com/sp/)	if you want	to validate documents as
###			they are being generated.
###	3.2.0 :	2003-05-02 : Morten Wulff <wulff@ratatosk.net>
###			Now handles subpages (directories named 2/ 3/ etc.).
###			Navigation between subpages must be in the page body as the
###			$self->{navigation} array is modified.
###	3.2.1 : 2003-05-03 : Morten Wulff <wulff@ratatosk.net>
###			Minor fix to prevent $self->{navigation} being clobbered on every
###			page.
### 3.2.2 : 2003-05-08 : Morten Wulff <wulff@ratatosk.net>
###			Added --update option
### 3.3.0 : 2003-05-12 : Morten Wulff <wulff@ratatosk.net>
###			Added support for external postprocessing subroutine
### 3.3.1 : 2003-05-25 : Morten Wulff <wulff@ratatosk.net>
###			Now displays all added or updated files (doesn't use cached modify
###			times)

### TODO
###			Update should delete files that have been removed from the cache.
###			Move configuration files to separate directory?

package Yggdrasil;

use strict;
use warnings;

$|++;

use AppConfig qw(:argcount);
use Carp;
use Date::Format;
use File::Copy;
use File::Find;
use File::Path;
use Getopt::Long;
use Text::Template qw(fill_in_string);
use Yggdrasil::Cache;
use Yggdrasil::Progress::Bar;

################################################################################
### INIT
################################################################################

our $AUTOLOAD;

### format of processing stats -- all valid %Fields% are used in this example
my $Stats_Template = <<'TEXT';

Processed {$files} files in {$directories} directories ({$total_size} kB, {$html_size} kB in html files)
Completed in {$elapsed_time} seconds

Changes:
{$changes}
Errors:
{$errors}
TEXT

### default settings - no trailing '/' on directory names and URLs
my %Fields = (
	include => [qw(\.html$ \.shtml$)],             ### include (process) files matching these regexp's, other files are just copied (include is applied after exclude)
	exclude => [qw(^_ \.tmpl$)],                   ### exclude files/paths matching these regexp's from processing
	skip    => [qw(^img ^index$)],                 ### exclude files/paths matching these regexp's from navigation

	input_dir  => '',                              ### location of source files (content and templates)
	output_dir => '',                              ### location of finished site
	data_dir   => 'data',                          ### store SDBM files here
	dir_mode   => 0644,                            ### dir create mode for mkpath

	base_url => 'http://www.psyke.org',            ### passed to templates (also used for 'permanent location' links)

	base_template        => 'base.tmpl',           ### base template - set to absolute path to override per-directory base templates
	template_extension   => 'tmpl',                ### used when looking for templates in the current path
	template_override    => undef,                 ### full path to template -- will be used instead of base.tmpl

	sp_command_line	=> '',						   ### {$filename} is replaced with	current	output filename

	home_label    => 'Home',                       ### name of the root page (used in breadcrumb trails)
	default_title => 'No Title',                   ### return for files not in database

	html_indices => [qw(index.html index.shtml)],  ### should match your apache config (are stripped from generated URL's)
	html_only    => 0,                             ### 1: give all html files the extension .html (useful when creating sites for offline use)

	front_page_navigation => 0,                    ### 1: create navigation array for top level of site
	include_index         => 0,                    ### 1: includes 'index.html' in navigation links (useful when creating sites for offline use)

	stats_template => $Stats_Template,             ### template to display site stats

	post_process_file  => undef,                   ### the name of a file containing an anonymous sub which will be used to postprocess all HTML files (see process())

	### the commandline switches override the following defaults
	arv         => '',        ### filename of an "A Real Validator" output file (<filename>.arv)
	config_file => 'default', ### filename of configuration file (<filename>.cfg) settings in configuration file overrides defaults
	debug       => 0,         ### 1: moderate debug info 2: more debug info 3: template info
	force       => 0,         ### force rebuild of cache (old database files are deleted)
	help        => 0,         ### 1: print usage
	update      => 0,         ### 1: only process files changed since last release, don't flush output directory
	verbose     => 0          ### level of verbosity
);

################################################################################
### PUBLIC METHODS
################################################################################

################################################################################
sub go {
	my $self = shift;

	### 'A Real Validator' output
	if ($self->{arv}) {
		my $arv = join '.', $self->{arv}, 'arv';
		open(ARV, ">$arv") or croak "Cannot open '$arv': $!\n";
		print ARV "\n";
		print "Will generate .arv file: '$arv'\n";
	}

	$self->_flush_output_directory() unless $self->{update};

	print "Processing '$self->{input_dir}':\n\n";

	### init progress bar
	$self->{progress_bar} = Yggdrasil::Progress::Bar->new($self->{cache}->size());

	### start processing directory tree
	find(
		{
			wanted      => $self->_bind_args(\&_process, $self),
			preprocess  => $self->_bind_args(\&_pre_process, $self),
			postprocess => $self->_bind_args(\&_post_process, $self)
		},
		$self->{input_dir}
	);

	close(ARV) if $self->{arv};
}

################################################################################
sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;

	my %parameters = @_;

	### initialize object
	my $self  = {
		_permitted => \%Fields
	};

	foreach my $element (keys %Fields) {
		$self->{$element} = $parameters{$element} || $Fields{$element};
	}

	bless ($self, $class);

	$self->_init(); ### complete object initialization

	return $self;
}

################################################################################
sub stats {
	my $self = shift;

	$self->{stats}->{total_size}   = int($self->{stats}->{total_size}/1024);
	$self->{stats}->{html_size}    = int($self->{stats}->{html_size}/1024);
	$self->{stats}->{elapsed_time} = $self->{stats}->{start_time} + time;

	return fill_in_string($self->{stats_template}, HASH	=> $self->{stats});
}

################################################################################
### PRIVATE METHODS
################################################################################

################################################################################
sub _pre_process {
	my $self  = shift;
	my @files = sort { $a cmp $b } @_;

	### remove unwanted elements from the file list
	@files = grep {
		my $file = $_;
		my $path = join '/', $File::Find::dir, $file;
		!/^\.{1,2}/ &&
		!grep {$file =~ /$_/ || $path =~ /$_/} @{$self->{exclude}}
	} @files;

	### update breadcrumb trail
	if ($#{$self->{trail}} > -1) {
		### handle sub-pages (/2 /3 etc.)
		my $name = $self->{cache}->title($File::Find::dir);
		if ($name eq $self->{trail}->[-1]->{name} && $File::Find::dir =~ /.*\/(\d+)$/) {
			$name = "Page $1";
		}

		push @{$self->{trail}}, {name => $name};
		$self->{stats}->{directories}++;
	}
	else {
		push @{$self->{trail}}, {name => $self->{home_label}};
	}

	### create links to subdirectories (except those in $self->{skip})
	if ($#{$self->{trail}} > 0 || $self->{front_page_navigation}) {
		@{$self->{navigation}} = map {
			{
				name => $self->{cache}->title(join '/', $File::Find::dir, $_),
				href => $self->{include_index} ? "$_/index.html" : "$_/"
			}
		} grep {
			my $file = $_;
			my $path = join '/', $File::Find::dir, $file;
			-d &&
			$self->{cache}->title($path) ne $self->{trail}->[-1]->{name} &&
			!grep {$file =~ /$_/ || $path =~ /$_/} @{$self->{skip}}
		} @files;
	}

	$self->{depth}++;

	return @files;
}

################################################################################
sub _process {
	my $self = shift;

	return if -d;

	$self->{progress_bar}->update() unless $self->{verbose};

	if ((stat($File::Find::name))[9] > $self->{release}) {
		$self->{stats}->{changes} .= "Updated: $File::Find::name\n";
		return if $self->{update};
	}

	### set target file
	my $target = $File::Find::name;
	$target =~ s/$self->{input_dir}/$self->{output_dir}/o;

	$self->_create_directory($target);

	### update stats
	my $file_size = (stat $File::Find::name)[7];
	$self->{stats}->{total_size} += $file_size;
	$self->{stats}->{files}++;

	### choose full processing or copy mode
	if ($self->_include($File::Find::name)) {
		print "HTML: " if $self->{debug};

		if ($self->{arv}) {
			my $filename = $target;
			$filename =~ s!/!\\!g;
			print ARV "$filename\n";
		}

		### modify trail if this isn't an index file
		my $non_index = 0;
		if (!$self->_html_index($File::Find::name)) {
			push @{$self->{trail}}, {name => $self->{cache}->title($File::Find::name)};
			$non_index = 1;
		}

		### get templates in current path (makes including templates easier)
		$self->_templates_in_path($File::Find::dir);

		### template override in effect
		$self->{template}->{__base__} = $self->{template_override} if $self->{template_override};

		### template stuff
		my $template = Text::Template->new(
			TYPE       => 'FILE',
			SOURCE     => $self->{template}->{__base__}
		) or croak "$Text::Template::ERROR";

		$target =~ s/(.*)\.(.*)/$1\.html/ if $self->{html_only};

		$self->{template}->{__file__} = $File::Find::name;
		$self->_update_content($File::Find::dir, $File::Find::name);

		open(OUT, ">$target") or croak "Cannot open '$target': $!\n";
		if ($self->{external_post_process}) {
			my $output = $template->fill_in(
				PACKAGE => 'Yggdrasil::Content'
			) or carp "$Text::Template::ERROR";
	        $output = &{$self->{external_post_process}}($output, $self->{exclude});
			print OUT $output;
		}
		else {
			$template->fill_in(
				PACKAGE => 'Yggdrasil::Content',
				OUTPUT => \*OUT
			) or carp "$Text::Template::ERROR";
		}
		close(OUT);

		$self->_validate_file($target) if $self->{validate};

		pop @{$self->{trail}} if $non_index;

		$self->{stats}->{html_size} += $file_size;
	}
	else {
		print "COPY: " if $self->{debug};
		copy($File::Find::name, $target) or croak "Cannot copy '$File::Find::name' to '$target': $!\n";
	}

	print "$File::Find::name\n" if $self->{debug} || $self->{verbose};
}

################################################################################
sub _post_process {
	my $self = shift;

	pop @{$self->{trail}};
	$self->{depth}--;
}



################################################################################
### http://perlmonks.thepen.com/109068.html
sub _bind_args {
	my $self = shift;
	my ($sub, @args) = @_;

	return sub {$sub->(@args, @_);};
}

################################################################################
sub _create_directory {
	my $self = shift;
	my $file = shift;

	my ($dir) = ($file =~ /(.*)\//);

	if (! -d $dir) {
		mkpath($dir , 0, $self->{dir_mode}) or croak "Cannot create directory '$dir': $!\n";
	}
}

################################################################################
sub _flush_output_directory {
	my $self = shift;

	print "Flushing output dir...";

	if (-d $self->{output_dir}) {
		rmtree($self->{output_dir}) or croak "Cannot remove directory '$self->{output_dir}': $!\n";
	}

	mkpath($self->{output_dir}, 0, $self->{dir_mode}) or croak "Cannot create directory '$self->{output_dir}': $!\n";

	print " Done!\n"
}

################################################################################
sub _get_release {
	my $self = shift;

	if (-e 'release.txt') {
		open(IN, 'release.txt') or croak "Cannot open 'release.txt': $!\n";
		my $temp = <IN>;
		close(IN);

		chomp $temp;
		$self->{release} = $temp;
	}
	else {
		print "'release.txt' doesn't exist. Using '0' as last update time...\n";
		$self->{release} = 0;
	}
}

################################################################################
sub _html_index {
	my $self = shift;
	my $file = shift;

	return grep {$file =~ /\/$_$/} @{$self->{html_indices}};
}

################################################################################
sub _include {
	my $self = shift;
	my $file = shift;

	return grep {$file =~ /$_/} @{$self->{include}};
}

################################################################################
sub _init {
	my $self = shift;

	### get command-line options
	GetOptions(
		'arv=s'    => \$self->{arv},
		'config=s' => \$self->{config_file},
		'debug'    => \$self->{debug},
		'force'    => \$self->{force},
		'help'     => \$self->{help},
		'update'   => \$self->{update},
		'validate' => \$self->{validate},
		'verbose'  => \$self->{verbose}
	);

	if ($self->{help}) {
		_print_usage();
		exit(0);
	}

	$self->_get_release();

	$self->_read_configuration_file();
	$self->_validate_configuration();

	### initialize stats counters
	$self->_init_stats();

	### initialize navigation
	$self->{trail}      = [];
	$self->{navigation} = [];

	$self->{depth} = -1;

	### initialize cache
	$self->{cache} = Yggdrasil::Cache->new(
		data_dir  => $self->{data_dir},
		input_dir => $self->{input_dir},
		force     => $self->{force},
		verbose   => $self->{verbose}
	);

	### initialize template utility functions
	*Yggdrasil::Content::include = sub {
		my $file = shift;
		Text::Template::fill_in_file($file, PACKAGE => 'Yggdrasil::Content');
	};
	*Yggdrasil::Content::include_text = \&Text::Template::_load_text;
}

################################################################################
sub _init_stats {
	my $self = shift;

	$self->{stats} = {
		files       => 0,
		directories => 0,
		total_size  => 0,
		html_size   => 0,
		start_time  => -time,
		changes     => '',
		errors		=> ''
	}
}

################################################################################
sub _print_usage {
	print <<TEXT;
Usage: y.pl [OPTION]...

  --arv=NAME     create output file NAME.arv for use with A Real Validator
  --config=NAME  use settings from NAME.cfg (will override defaults in y.pl)
  --debug        print debug information
  --force        force rebuild of cache
  --help         print this message
  --verbose      print directory names and other information while processing
TEXT
}

################################################################################
sub _read_configuration_file {
	my $self = shift;
	my $file = join '.', $self->{config_file}, 'cfg';

	unless (-e $file) {
		carp "Configuration file $file doesn't exist";
		return;
	}

	### initialize AppConfig
	my $config = AppConfig->new({CASE=>1});
	foreach my $key (keys %Fields) {
		$config->define(
			$key => ref($Fields{$key}) eq 'ARRAY'
				  ? {ARGCOUNT => ARGCOUNT_LIST}
				  : {ARGCOUNT => ARGCOUNT_ONE}
		);
	}

	### read configuration file
	$config->file($file);

	### settings from the configuration file override the defaults in $self
	my %varlist = $config->varlist('.*');
	foreach my $key (keys %varlist)
	{
		if (ref($config->get($key)) eq 'ARRAY') {
			$self->{$key} = $config->get($key) if $#{$config->get($key)} > -1;
		}
		else {
			$self->{$key} = $config->get($key) if $config->get($key);
		}
	}

	### create new data directories for alternative configuratons
	if ($self->{config_file} ne 'default') {
		$self->{data_dir} = join '/', $self->{data_dir}, $self->{config_file};
	}

	### initialize external post processing if present
	my $sub_file = join '.', $self->{config_file}, 'sub';
	if (-e $sub_file) {
		$self->{external_post_process} = do($sub_file) or die "Cannot init external post processing '$sub_file': $!\n";
	}
}

################################################################################
sub _templates_in_path {
	my $self = shift;
	my $dir  = shift;

	my @path = split /\//, $dir;

	$self->{template} = {};

	while (@path) {
		my $dir = join '/', @path;

		foreach my $template_path (glob("$dir/*.$self->{template_extension}")) {
			my ($template_name) = ($template_path =~ /.*\/(.*)/);
			$self->{template}->{$template_name} ||= $template_path;
		}

###		last if $dir eq $self->{input_dir};
		pop @path;
	}

	$self->{template}->{__base__} = $self->{template}->{$self->{base_template}} or croak "Base template not found in path";
}

################################################################################
sub _trail_links {
	my $self = shift;

	my $dots = $#{$self->{trail}};

	foreach my $item (@{$self->{trail}}) {
		$item->{href}  = $dots ? '../' x $dots : '';
		$item->{href} .= 'index.html' if ($self->{include_index} && $dots);
		$dots--;
	}

	### handle trails for non-index html files
	unless ($self->_html_index($File::Find::name)) {
		my $index_pos = $#{$self->{trail}}-1;
		my $count = 0;
		foreach (@{$self->{trail}}) {
			if ($count == $index_pos) {
				$_->{href} = './';
			} else {
				$_->{href} =~ s!../!!;
			}
			$count++;
		}
	}
}

################################################################################
sub _update_content {
	my $self = shift;
	my ($dir, $file) = @_;

	$self->_trail_links();

	my $location = $dir;
	$location =~ s/$self->{input_dir}/$self->{base_url}/o;

	### make sure that permanent location works for non-index files
	if ($file !~ /index/) {
		my ($filename) = ($file =~ /.*\/(.*)/);
		$location .= "/$filename" ;
	}

	$Yggdrasil::Content::title = $self->{cache}->title($file);

	@Yggdrasil::Content::trail      = @{$self->{trail}};
	@Yggdrasil::Content::navigation = @{$self->{navigation}};
	$Yggdrasil::Content::depth      = $self->{depth};
	$Yggdrasil::Content::location   = $location;
	$Yggdrasil::Content::rel_path   = '../' x ( $self->{depth} );

	%Yggdrasil::Content::template   = %{$self->{template}};

	$Yggdrasil::Content::timestamp  = time2str('%a %b %e %Y (%T)', time);
}

################################################################################
sub _validate_configuration {
	my $self = shift;

	carp "No file inclusion regular expressions specified" unless $#{$self->{include}} >= 0;
	carp "No file exclusion regular expressions specified" unless $#{$self->{exclude}} >= 0;
	carp "Input directory '$self->{input_dir}' doesn't exist" unless -d $self->{input_dir};
	carp "No base URL specified" unless $self->{base_url};
	carp "No base template specified" unless $self->{base_template};
	carp "No HTML indices specified" unless $#{$self->{html_indices}} >= 0;
}

################################################################################
sub _validate_file	{
	my $self = shift;
	my $file = shift;

	my $validate = fill_in_string($self->{sp_command_line},	HASH =>	{filename => $file});

	my @result = split/\n+/, `$validate 2>&1`;
	foreach	(@result) {
		next unless /$file/;
		s/(.*?)$file://;
		my ($line, $char, $type, $message) = split /:/;

		### message can be empty
		$message =~ s/^\s+// if $message;
		$message ||= '';

		$self->{stats}->{errors} .= "$file\n\t($line, $char)\t$message\n";
	}
}

################################################################################
sub AUTOLOAD {
	my $self = shift;
	my $type = ref($self) or croak "$self is not an object";

	my $name = $AUTOLOAD;
	$name =~ s/.*://;

	unless (exists $self->{_permitted}->{$name} ) {
		croak "Can't access '$name' field in class $type";
	}

	if (@_) {
		return $self->{$name} = shift;
	}
	else {
		return $self->{$name};
	}
}

1;

################################################################################
### DOCUMENTATION
################################################################################

__END__

=head1 NAME

Yggdrasil - Generates a website from a directory tree of content and templates

=head1 SYNOPSIS

You can either pass the configuration when creating a new Yggdrasil object or
use a configuration file (if you use both, the configuratio file will take
precedence).

=head1 DESCRIPTION


=head2 Configuration variables

=over

=item include

Files matching one or more of the regular expressions in this array are
processed by Text::Template. Files that do not match any of the regular
expressions are copied to the target directory without processing. Should only
match HTML files.

=item exclude

Files or directories matching one or more of the regular expressions in this
array are not processed or copied. These regular expressions are applied before
the regular expressions in the include array.

=item skip

Directories matching one or more of the regular expressions in this array are
not included in the navigation array (see below).

=item input_dir

Path to source directory.

=item input_dir

Path to output directory.

=item data_dir

Directory used to store title and update time caches.

=item dir_mode

Directory create mode used by mkpath().

=item base_url

Passed to templates.

=item base_template

Filename of base template. All parent directories of the current directory are
searched.

=item template_extension

File extension of template files.

=item template_override

This template will be used instead of the templates in the source directory.

=item sp_command_line

Set this to the complete commandline for nsgmls if you wish to validate output
files while processing. {$filename} is replaced with the current filename. You
need to have the SP SGML system (http://www.jclark.com/sp/) installed in order
to use this feature.

=item home_label

Name given to the root page in the breadcrumb trail.

=item default_title

Title given to files not found in the title cache.

=item html_indices

These filenames are removed from URL's in the breadcrumb trail. Should match
your webserver configuration.

=item html_only

Set to 1 to give all HTML files (the files in include) the extension .html
(this is useful when creating site versions for offline use or for webservers
that can't handle SSI).

=item front_page_navigation

Set to 1 to create a navigation array for the top level of the site (useful when
creating site versions for offline use).

=item include_index

Set to 1 to include 'index.html' in navigation and breadcrumb trail links
(useful when creating site versions for offline use).

=item stats_template

Template used to display site stats.

=item post_process_file

Name of file containing an anonymous sub. The sub will be called once for each
HTML file in the site.

=back


=head2 Template variables

=over

=item $depth

Number of directory levels from the current page to the root level. Can be used
to include or exclude template elements depending on the position of the current
page in the site structure.

=item $location

The URL of the current page.

=item $rel_path

The path from the current page to the root page.

=item $timestamp

Time of template initialization.

=item $title

Page title.

=item {@navigation}

Contains links to subpages of the current page. Each element is a hashref of the
form

    {
        name => <page title>,
        href => <relative url>
    }

=item {@trail}

First item...

=back


=head2 Template functions

=over

=item include

First item...

=item include_text

First item...

=back


=head1 AUTHOR

Morten Wulff, <wulff@psyke.org>


=head1 BUGS

No known bugs.


=head1 COPYRIGHT

Copyright (C) 2003 Morten Wulff

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=head1 VERSION

Version 3.3.1
